# Wendelin.bigfile | common ZODB-related helpers
# Copyright (C) 2014-2020  Nexedi SA and Contributors.
#                          Kirill Smelkov <kirr@nexedi.com>
#
# This program is free software: you can Use, Study, Modify and Redistribute
# it under the terms of the GNU General Public License version 3, or (at your
# option) any later version, as published by the Free Software Foundation.
#
# You can also Link and Combine this program with other software covered by
# the terms of any of the Free Software licenses or any of the Open Source
# Initiative approved licenses and Convey the resulting work. Corresponding
# source of such a combination shall include the source code for all other
# software used.
#
# This program is distributed WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See COPYING file for full licensing terms.
# See https://www.nexedi.com/licensing for rationale and options.
"""Package wendelin.lib.zodb provides ZODB-related utilities."""

import ZODB
from ZODB.FileStorage import FileStorage
from ZODB import DB
from ZODB import POSException
from ZODB.utils import p64, u64
from persistent import Persistent
from weakref import WeakSet
import gc
from persistent.TimeStamp import TimeStamp
import time
from ZODB.utils import z64

import pkg_resources


# open db storage by uri
def dbstoropen(uri):
    # if we can - use zodbtools to open via zodburi
    try:
        import zodbtools.util
    except ImportError:
        return _dbstoropen(uri)

    return zodbtools.util.storageFromURL(uri)


# simplified fallback to open a storage by URI when zodbtools/zodburi are not available.
# ( they require ZODB, instead of ZODB3, and thus we cannot use
#   them together with ZODB 3.10 which we still support )
def _dbstoropen(uri):
    if uri.startswith('neo://'):
        # XXX hacky, only 1 master supported
        from neo.client.Storage import Storage
        name, master = uri[6:].split('@', 1)    # neo://db@master -> db, master
        stor = Storage(master_nodes=master, name=name)

    elif uri.startswith('zeo://'):
        # XXX hacky
        from ZEO.ClientStorage import ClientStorage
        host, port = uri[6:].split(':',1)       # zeo://host:port -> host, port
        port = int(port)
        stor = ClientStorage((host, port))

    else:
        stor = FileStorage(uri)

    return stor


# open stor/db/connection and return root obj
def dbopen(uri):
    stor = dbstoropen(uri)
    db   = DB(stor)
    conn = db.open()
    root = conn.root()
    return root


# close db/connection/storage identified by root obj
def dbclose(root):
    conn = root._p_jar
    db   = conn.db()
    stor = db.storage

    conn.close()
    db.close()
    stor.close()


# LivePersistent is Persistent that never goes to ghost state, if it was ever uptodate.
#
# NOTE
#
# On invalidation LivePersistent still goes to ghost state, because
# invalidation cannot be ignored, i.e. they indicate the object has been
# changed externally.
#
# Invalidation can happen only at transaction boundary, so during the course of
# transaction LivePersistent is guaranteed to stay uptodate.
class LivePersistent(Persistent):
    # don't allow us to go to ghost
    #
    # NOTE we can't use STICKY as that state is assumed as
    # short-lived-temporary by ZODB and is changed back to UPTODATE by
    # persistent code. In fact ZODB says: STICKY is UPTODATE+keep in memory.
    def _p_deactivate(self):
        # just returning here won't allow Persistent._p_deactivate() run and
        # thus we'll stay in non-ghost state.
        return

    # NOTE _p_invalidate() is triggered on invalidations. We do not override it.



# deactivate a btree, including all internal buckets and leaf nodes
def deactivate_btree(btree):
    # first activate btree, to make sure its first bucket is loaded at all.
    #
    # we have to do this because btree could be automatically deactivated
    # before by cache (the usual way) and then in its ghost state it does not
    # contain pointer to first bucket and thus we won't be able to start
    # bucket deactivation traversal.
    btree._p_activate()

    for _ in gc.get_referents(btree):
        # for top-level btree we ignore any direct referent besides bucket
        # (there are _p_jar, cache, etc)
        if type(_) is btree._bucket_type:
            _deactivate_bucket(_)

    btree._p_deactivate()

def _deactivate_bucket(bucket):
    # TODO also support objects in keys, when we need it
    for obj in bucket.values():
        if type(obj) == type(bucket):
            _deactivate_bucket(obj)
        elif isinstance(obj, Persistent):
            obj._p_deactivate()

    bucket._p_deactivate()


# zconn_at returns tid as of which ZODB connection is viewing the database.
def zconn_at(zconn): # -> tid
    assert isinstance(zconn, ZODB.Connection.Connection)
    if zconn.opened is None: # zconn must be in "opened" state
        raise POSException.ConnectionStateError("database connection is closed")
    # ZODB5 uses MVCC uniformly
    #
    # zconn.db._storage always provides IMVCCStorage - either raw storage provides it,
    # or DB wraps raw storage with MVCCAdapter.
    #
    # MVCCAdapter in turn uses either MVCCAdapterInstance (current) or
    # HistoricalStorageAdapter, or UndoAdapterInstance. Retriving zconn.at from those:
    #
    # MVCCAdapterInstance
    #     ._start
    #
    # HistoricalStorageAdapter
    #     ._before
    #
    # UndoAdapterInstance
    #     # no way to retrieve `at`, but .undo_instance() through which
    #     # UndoAdapterInstance is returnerd, is not used anywhere.
    #
    # For the reference: FileStorage, ZEO and NEO do not provide IMVCCStorage, thus
    # for them we can rely on MVCCAdapterInstance.
    #
    # RelStorage is IMVCCStorage - TODO: how to extract at.
    if zmajor >= 5:
        zstor = zconn._storage
        if isinstance(zstor, ZODB.mvccadapter.MVCCAdapterInstance):
            return before2at(zstor._start)

        if isinstance(zstor, ZODB.mvccadapter.HistoricalStorageAdapter):
            return before2at(zstor._before)

        raise AssertionError("zconn_at: TODO: add support for zstor %r" % zstor)

    elif zmajor == 4:
        if zconn.before is None:
            zconn._inv_lock.acquire()
            try:
                viewed_state = zconn.viewed_db_state
            finally:
                zconn._inv_lock.release()
            return viewed_state
        else:
            return before2at(zconn.before)
    
    # TODO ZODB3  
    else: 
        pass

# before2at converts tid that specifies database state as "before" into tid that
# specifies database state as "at".
def before2at(before): # -> at
    return p64(u64(before) - 1)


# _zversion returns ZODB version object
def _zversion():
    dzodb3 = pkg_resources.working_set.find(pkg_resources.Requirement.parse('ZODB3'))
    dzodb  = pkg_resources.working_set.find(pkg_resources.Requirement.parse('ZODB'))
    v311   = pkg_resources.parse_version('3.11dev')

    if dzodb3 is None and dzodb is None:
        raise RuntimeError('ZODB is not installed')

    if dzodb3 is not None:
        if dzodb3.parsed_version >= v311:
            vzodb = dzodb.parsed_version # ZODB 3.11 just requires latest ZODB & ZEO
        else:
            vzodb = dzodb3.parsed_version
    else:
        vzodb = dzodb.parsed_version

    assert vzodb is not None
    return vzodb

# _zmajor returns major ZODB version.
def _zmajor():
    vzodb = _zversion()
    # XXX hack - packaging.version.Version provides no way to extract major?
    return int(vzodb.public.split('.')[0])  # 3.11.dev0 -> 3

# zmajor is set to major ZODB version.
zmajor = _zmajor()


# patch for ZODB.Connection to support callback on .open()
# NOTE on-open  callbacks are setup once and fire many times on every open
#      on-close callbacks are setup once and fire only once on next close
ZODB.Connection.Connection._onOpenCallbacks = None
def Connection_onOpenCallback(self, f):
    if self._onOpenCallbacks is None:
        # NOTE WeakSet does not work for bound methods - they are always created
        # anew for each obj.method access, and thus will go away almost immediately
        self._onOpenCallbacks = WeakSet()
    self._onOpenCallbacks.add(f)

assert not hasattr(ZODB.Connection.Connection, 'onOpenCallback')
ZODB.Connection.Connection.onOpenCallback = Connection_onOpenCallback

_orig_Connection_open = ZODB.Connection.Connection.open
def Connection_open(self, transaction_manager=None, delegate=True):
    _orig_Connection_open(self, transaction_manager, delegate)

    # FIXME method name hardcoded. Better not do it and allow f to be general
    # callable, but that does not work with bound method - see above.
    # ( Something like WeakMethod from py3 could help )
    if self._onOpenCallbacks:
        for f in self._onOpenCallbacks:
            f.on_connection_open()

ZODB.Connection.Connection.open = Connection_open


# patch for ZODB.Connection to support callback on after database view is changed
ZODB.Connection.Connection._onResyncCallbacks = None
def Connection_onResyncCallback(self, f):
    if zmajor <= 4:
        raise AssertionError("onResyncCallback: TODO: add support for ZODB34")
    if self._onResyncCallbacks is None:
        # NOTE WeakSet does not work for bound methods - they are always created
        # anew for each obj.method access, and thus will go away almost immediately
        self._onResyncCallbacks = WeakSet()
    self._onResyncCallbacks.add(f)

assert not hasattr(ZODB.Connection.Connection, 'onResyncCallback')
ZODB.Connection.Connection.onResyncCallback = Connection_onResyncCallback

# ZODB5: hook into Connection.newTransaction
if zmajor >= 5:
    _orig_Connection_newTransaction = ZODB.Connection.Connection.newTransaction
    def _ZConnection_newTransaction(self, transaction, sync=True):
        _orig_Connection_newTransaction(self, transaction, sync)

        # FIXME method name hardcoded. Better not do it and allow f to be general
        # callable, but that does not work with bound method - see above.
        # ( Something like WeakMethod from py3 could help )
        if self._onResyncCallbacks:
            for f in self._onResyncCallbacks:
                f.on_connection_resync()

    ZODB.Connection.Connection.newTransaction = _ZConnection_newTransaction


# ZODB4: hook into Connection._storage_sync
elif zmajor == 4:
    pass    # raises in onResyncCallback


# ZODB3: TODO
else:
    pass    # raises in onResyncCallback


if zmajor == 4:
    orig_connection_constructor = ZODB.Connection.Connection.__init__
    def connection_constructor(self, db, cache_size=400, before=None, cache_size_bytes=0):
        orig_connection_constructor(self, db, cache_size, before, cache_size_bytes)
        
        # viewed state of the db
        self.viewed_db_state = self._db.latest_state
    ZODB.Connection.Connection.__init__ = connection_constructor

    orig_db_constructor = ZODB.DB.__init__
    def db_constructor(self, storage,
                 pool_size=7,
                 pool_timeout=1<<31,
                 cache_size=400,
                 cache_size_bytes=0,
                 historical_pool_size=3,
                 historical_cache_size=1000,
                 historical_cache_size_bytes=0,
                 historical_timeout=300,
                 database_name='unnamed',
                 databases=None,
                 xrefs=True,
                 large_record_size=1<<24,
                 **storage_args):

        orig_db_constructor(self, storage,
                    pool_size,
                    pool_timeout,
                    cache_size,
                    cache_size_bytes,
                    historical_pool_size,
                    historical_cache_size,
                    historical_cache_size_bytes,
                    historical_timeout,
                    database_name,
                    databases,
                    xrefs,
                    large_record_size,
                    **storage_args)
        
        # initialize viewed state of db for new connections being opened
        last_tid = self.lastTransaction()
        self.history(z64)
        self._a()
        try:
            # initialize latest_state with last_tid if not yet initialized
            # during invalidation process
            if not hasattr(self, 'latest_state'):
                self.latest_state = last_tid
        finally:
            self._r()

    ZODB.DB.__init__= db_constructor

    def db_connectionMap(self, f, tid=None):
        """Call f(c) for all connections c in all pools, live and historical.
        """
        self._a()
        try:
            # updating latest_state for zconn_at
            if not (tid is None):
                self.latest_state = tid

            self.pool.map(f)
            self.historical_pool.map(f)
        finally:
            self._r()
    
    ZODB.DB._connectionMap = db_connectionMap

    def db_invalidate(self, tid, oids, connection=None, version=''):
        """Invalidate references to a given oid.

        This is used to indicate that one of the connections has committed a
        change to the object.  The connection commiting the change should be
        passed in to prevent useless (but harmless) messages to the
        connection.
        """
        # Storages, esp. ZEO tests, need the version argument still. :-/
        assert version==''
        # Notify connections.
        def inval(c):
            if c is not connection:
                c.invalidate(tid, oids)
            else:
                # update resynced connection's view of db
                c._inv_lock.acquire()
                try:
                    c.viewed_db_state = tid
                finally:
                    c._inv_lock.release()

        self._connectionMap(inval, tid)
    
    ZODB.DB.invalidate = db_invalidate

    def connection_flush_invalidations(self):
        if self._mvcc_storage:
            # Poll the storage for invalidations.
            invalidated = self._storage.poll_invalidations()
            if invalidated is None:
                # special value: the transaction is so old that
                # we need to flush the whole cache.
                self._cache.invalidate(list(self._cache.cache_data.keys()))
            elif invalidated:
                self._cache.invalidate(invalidated)
        
        self._db._a()
        self._inv_lock.acquire()
        try:
            # Non-ghostifiable objects may need to read when they are
            # invalidated, so we'll quickly just replace the
            # invalidating dict with a new one.  We'll then process
            # the invalidations after freeing the lock *and* after
            # resetting the time.  This means that invalidations will
            # happen after the start of the transactions.  They are
            # subject to conflict errors and to reading old data.

            # TODO: There is a potential problem lurking for persistent
            # classes.  Suppose we have an invalidation of a persistent
            # class and of an instance.  If the instance is
            # invalidated first and if the invalidation logic uses
            # data read from the class, then the invalidation could
            # be performed with stale data.  Or, suppose that there
            # are instances of the class that are freed as a result of
            # invalidating some object.  Perhaps code in their __del__
            # uses class data.  Really, the only way to properly fix
            # this is to, in fact, make classes ghostifiable.  Then
            # we'd have to reimplement attribute lookup to check the
            # class state and, if necessary, activate the class.  It's
            # much worse than that though, because we'd also need to
            # deal with slots.  When a class is ghostified, we'd need
            # to replace all of the slot operations with versions that
            # reloaded the object when called. It's hard to say which
            # is better or worse.  For now, it seems the risk of
            # using a class while objects are being invalidated seems
            # small enough to be acceptable.

            invalidated = dict.fromkeys(self._invalidated)
            self._invalidated = set()
            self._txn_time = None
            if self._invalidatedCache:
                self._invalidatedCache = False
                invalidated = self._cache.cache_data.copy()
            # update resynced connection's view of the db
            self.viewed_db_state = self._db.latest_state
        finally:
            self._inv_lock.release()
            self._db._r()

        self._cache.invalidate(invalidated)

        # Now is a good time to collect some garbage.
        self._cache.incrgc()
    
    ZODB.Connection.Connection._flush_invalidations = connection_flush_invalidations

# zstor_2zurl converts a ZODB storage to URL to access it.
def zstor_2zurl(zstor):
    # There is, sadly, no unified way to do it, as even if storages are created via
    # zodburi, after creation its uri is lost. And storages could be created not
    # only through URI but e.g. via ZConfig and manually. We want to support all
    # those cases...
    #
    # For this reason extract URL with important for wcfs use-case parameters in
    # ad-hoc way.
    if isinstance(zstor, FileStorage):
        return "file://%s" % (zstor._file_name,)

    # TODO ZEO + NEO support
    raise NotImplementedError("don't know how to extract zurl from %r" % zstor)
